#!/usr/bin/env ruby

# == Synopsis 
#   This is the driver to run the benchmarks for the Sequentially Consistent Java.
#
# == Examples
#   This command is a driver for the benchmarks.
#     run_benchmark 
#
# == Usage 
#   run_benchmark [options] benchmark_prefix [benchmark_size]
#   benchmark_size        'small' or 'large' if the benchmark accepts input
#
#   For help use: run_benchmark -h
#
# == Options
#   -h, --help          Displays help message
#   -q, --quiet         Output as little as possible, overrides verbose
#   -V, --verbose       Verbose output
#   -t                  Time execution
#   -server            Start java with the -server flag
#   -ea                 use -ea flag for java
#   TO DO - add additional options
#
# == Author
#   Christoph Angerer
#
# == Copyright
#   Copyright (c) 2011 Christoph Angerer. Licensed under the GNU Lesser General Public License:
#   http://www.gnu.org/licenses/lgpl.html

require 'optparse'
require 'rdoc/usage'
require 'ostruct'
require 'date'

class App

  attr_reader :options

  def initialize(arguments, stdin)
    @arguments = arguments
    @stdin = stdin

    #set defaults
    @options = OpenStruct.new
    @options.verbose = false
    @options.quiet = false
  end

  def run
    if parsed_options? && arguments_valid?
      puts "Start at #{DateTime.now}" if @options.verbose
  
      output_options if @options.verbose
  
      process_arguments
      process_command
  
      puts "Finished at #{DateTime.now}" if @options.verbose
    else
      output_usage
    end
  end

  protected
  def parsed_options?
    opts = OptionParser.new
    opts.on('-h', '--help')       { output_help }
    opts.on('-V', '--verbose')    { @options.verbose = true }  
    opts.on('-q', '--quiet')      { @options.quiet = true }
    opts.on('-t')                 { @options.time = true }
    opts.on('-server') { @options.server = true }
    opts.on('-ea') { @options.ea = true }
    
    opts.parse!(@arguments) rescue return false
  
    process_options
    true      
  end

  # Performs post-parse processing on options
  def process_options
    @options.verbose = false if @options.quiet
  end

  def output_options
    puts "Options:"
  
    @options.marshal_dump.each do |name, val|        
      puts "  #{name} = #{val}"
    end
  end

  # True if required arguments were provided
  def arguments_valid?
    # TO DO - implement your real logic here
    return false if @arguments.length < 1 || @arguments.length > 2    
    true
  end

  # Setup the arguments
  def process_arguments
    # mapping from benchmark to its main class and input params based on input size
  @benchmarks = {
    :philo => {:main => "philo.top.philo.Philo"},
    :sor => {:main => "sor.top.sor.Sor", :large => "10000 4"},
    :tsp => {:main => "tsp.top.tsp.Tsp", :large => "Shared/benchmarks/Erco/tsp/tspfiles/tspfile20.large"},
    :hedc => {:main => "hedc.top.hedc.Tester"},
    :elevator => {:main => "elevator.top.elevator.Elevator", :large => "bin/elevator/data3"},
    :dummy => nil
  }
  
    bench = @arguments[0]
    size = @arguments[1] || "large"
    config = @benchmarks[bench.to_sym]
    @mainclass = config[:main]
    @input = config[size.to_sym] || ""
  end

  def output_help
    RDoc::usage() #exits app
  end

  def output_usage
    RDoc::usage('usage') # gets usage from comments above
  end

  def process_command
    bootclasspath = "-classpath ./bin/:./Shared/libs/jsr166y.jar:../Shared/libs/javassist.jar"
    java_command = "java #{"-server" if @options.server} #{"-ea" if @options.ea } -javaagent:../Shared/libs/top_runtime.jar -Xmx1024m #{bootclasspath} #{@mainclass} #{@input}"
    
    command = "#{"time " if @options.time}#{java_command}"
    puts command
    result = `#{command}`
    puts result
  end
end

app = App.new(ARGV, STDIN)
app.run